//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

//#define ENVIRONMENT												// Add environment mapping effect
//#define LIGHTMAP													// Add lightmap
//#define LIGHTMAP_ONLY												// Display only lightmap
//#define SKIN														// Compute position based on bones
//#define FOG														// Add fog
//#define DISABLE_BUMP_AND_PARALLAX									// Only diffuse color from texture
//#define DISABLE_PARALLAX											// Bump mapping without parallax
//#define OUTPUT_DEPTH												// Output depth in second render target
//#define USE_MATERIAL_COLOR										// Use material color instead of diffuse texture (use with DISABLE_BUMP_AND_PARALLAX)
//#define COMPRESSED_VECTORS										// Compressed vectors

#if USE_MATERIAL_COLOR
#define DISABLE_BUMP_AND_PARALLAX 1
#define DISABLE_PARALLAX 1
#endif

#if DISABLE_BUMP_AND_PARALLAX
#define DISABLE_PARALLAX 1
#endif

float4x4	g_mWorldViewProjection;									// World * View * Projection matrix
float4		g_LightPositionOS;										// Light position in object space
float4		g_EyePositionOS;										// Camera position in object space

#if !USE_MATERIAL_COLOR
texture		g_BaseTexture;											// RGB = color, A = glow strength
#else
float4		g_MaterialColor = float4( 1.0f, 1.0f, 1.0f, 1.0f );	// Material color
#endif

#if !DISABLE_BUMP_AND_PARALLAX
texture		g_BumpTexture;											// RGB = normal, A = specular power
texture		g_HeightTexture;										// R = height, G = environment mapping strength
#endif

#if LIGHTMAP
texture		g_LightmapTexture;
float4		g_vLightMapParam = float4( 1.5f, 0.2f, 0.0f, 0.0f );	// lightmap_color = lightmap_color * X + Y
#endif

#if ENVIRONMENT
textureCUBE	g_EnvironmentTexture;
#endif

#if SKIN
static const int MAX_MATRICES = 26;
float4x4	g_mWorldMatrixArray[ MAX_MATRICES ];
float4x4	g_mViewProj;
#define		NUM_OF_BONES 2
#endif

#if FOG
float4		g_vFogColor;											// Fog color
float4		g_vFogParams;											// X = fog near plane, Y = fog far plane (X < Y), Z = camera far clip plane distance
#define		FOG_NEAR_PLANE				0
#define		FOG_FAR_PLANE				1
#define		FOG_CAMERA_FAR_CLIP_PLANE	2
#endif

#if OUTPUT_DEPTH
#include "ComputeDepth.fx"
float			g_fCameraFarPlaneDist = 1500.0f;					// Distance to the far clip plane (the same as g_vFogParams.y)
#define 		DEPTH_INDEX			3
#endif

#if LIGHTMAP_ONLY && !LIGHTMAP
"ERROR - you need to enable lightmaps first"
#endif

#if LIGHTMAP_ONLY && OUTPUT_DEPTH
"ERROR - cannot output only lightmap color and depth"
#endif

#if ENVIRONMENT && DISABLE_BUMP_AND_PARALLAX
"ERROR - cannot mix envoronment mapping without bump and parallax mapping"
#endif

#if USE_MATERIAL_COLOR && !DISABLE_BUMP_AND_PARALLAX
"ERROR - disable bump and parallax if we are not going to use base texture"
#endif

#if USE_MATERIAL_COLOR && ( LIGHTMAP || LIGHTMAP_ONLY )
"ERROR - cannot mix material color with lightmap"
#endif

#if USE_MATERIAL_COLOR && ENVIRONMENT
"ERROR - cannot mix material color with environment mappings"
#endif

#if !COMPRESSED_VECTORS
"ERROR - only compressed vectors"
#endif

//-----------------------------------------------------------------------------
// Texture sampler
//-----------------------------------------------------------------------------
#if !USE_MATERIAL_COLOR
sampler BaseTextureSampler = sampler_state
{
	Texture		= < g_BaseTexture >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
};
#endif

#if !DISABLE_BUMP_AND_PARALLAX
sampler BumpTextureSampler = sampler_state
{
	Texture		= < g_BumpTexture >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
};
#endif

#if !DISABLE_PARALLAX || ENVIRONMENT
sampler HeightTextureSampler = sampler_state
{
	Texture		= < g_HeightTexture >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
};
#endif

#if ENVIRONMENT
sampler EnvironmentMapSampler = sampler_state
{
	Texture		= < g_EnvironmentTexture >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
    AddressW	= WRAP;
};
#endif

#if LIGHTMAP
sampler LightMapSampler = sampler_state
{
	Texture		= < g_LightmapTexture >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
};
#endif

float3 expand(float3 v)
{
    return (v-0.5)*2;
}

//**********************************************************************************************
// Factory
//**********************************************************************************************

void VS_Factory(	float4 i_Position						: POSITION,
					float3 i_Normal							: NORMAL,
#if SKIN
					float4  i_BlendWeights					: BLENDWEIGHT,
					float4  i_BlendIndices					: BLENDINDICES,
#endif
#if !USE_MATERIAL_COLOR
					float2 texCoord							: TEXCOORD0,
#endif
#if LIGHTMAP
					float2 lighmapTexCoord					: TEXCOORD1,
#endif
#if !DISABLE_BUMP_AND_PARALLAX
					float3 i_Tangent						: TANGENT,
#endif
					out float4 oPosition					: POSITION,
#if !LIGHTMAP
#if !USE_MATERIAL_COLOR
					out float2 oTextureCoord				: TEXCOORD0,
#endif					
#else
					out float4 oTextureCoord				: TEXCOORD0,
#endif
#if !OUTPUT_DEPTH
					out float3 oLightDirectionAndDepth		: TEXCOORD1,
#else
					out float4 oLightDirectionAndDepth		: TEXCOORD1,
#endif
#if !DISABLE_BUMP_AND_PARALLAX                    
					out float3 oHalfVector					: TEXCOORD2,					
#if !DISABLE_PARALLAX					
					out float3 oEyeVector					: TEXCOORD3,
#endif					
#else                    
					out float3 oNormalOS					: TEXCOORD2,
#endif
#if ENVIRONMENT
					out float3 oReflectedVectorWS			: TEXCOORD4,
#endif
#if FOG
					out float oPixelDepth					: TEXCOORD5,
#endif
					uniform float3   lightPosition,  // Object-space
					uniform float3   eyePosition,    // Object-space
					uniform float4x4 modelViewProj )
{
#if !SKIN
    
	// Transform the position from object space to homogeneous projection space
	oPosition = mul( i_Position, modelViewProj );
    
#else
    
	// NOTE: This shader is invalid, because it doesn't compute proper normal vector and object space position,
	//       but in this case, it is not a problem - skinned mesh is very thin and it looks good even with this "bug"
    
	float4		vPositionWorld	= 0.0f;
	float		LastWeight		= 0.0f;
     
	// Cast the vectors to arrays for use in the for loop below
	float	BlendWeightsArray[4]	= (float[4])i_BlendWeights;  
	int		IndexArray[4]			= (int[4])i_BlendIndices;

	// Calculate the pos/normal using the "normal" weights 
	// and accumulate the weights to calculate the last weight
	for (int iBone = 0; iBone < NUM_OF_BONES-1; iBone++)
	{
		LastWeight = LastWeight + BlendWeightsArray[iBone];
		vPositionWorld += mul( i_Position, g_mWorldMatrixArray[IndexArray[iBone]] ) * BlendWeightsArray[iBone];
	}
	LastWeight = 1.0f - LastWeight; 

	// Now that we have the calculated weight, add in the final influence
	vPositionWorld += ( mul(i_Position, g_mWorldMatrixArray[IndexArray[NUM_OF_BONES-1]]) * LastWeight );

	// Transform position from world space into view and then projection space
	oPosition = mul( vPositionWorld, g_mViewProj );

#endif    
    
#if COMPRESSED_VECTORS
	// Unpack normal vector
	i_Normal = expand( i_Normal );

#if !DISABLE_BUMP_AND_PARALLAX
	// Unpack normal vector
	i_Tangent = expand( i_Tangent );
#endif
	
#endif

#if !USE_MATERIAL_COLOR
	// Texture coordinates
	oTextureCoord.xy = texCoord.xy;
#endif	

	// Compute the object-space light vector
	oLightDirectionAndDepth.xyz = lightPosition - i_Position.xyz;    
    
#if !DISABLE_BUMP_AND_PARALLAX
	// Construct object space to texture space 3x3 matrix
	float3 binormal = cross( i_Tangent, i_Normal );
	float3x3 rotation = transpose( float3x3( i_Tangent, binormal, i_Normal ) );

	float3 vCameraDirection = eyePosition - i_Position.xyz;
	oHalfVector = oLightDirectionAndDepth.xyz + vCameraDirection;

	// Rotate the light and half vector using the matrix
	oLightDirectionAndDepth.xyz = mul( oLightDirectionAndDepth.xyz, rotation );
	oHalfVector = mul( oHalfVector, rotation );
#endif
    
#if !DISABLE_PARALLAX
	// Eye vector for parallax mapping
	float3x3 T = float3x3( binormal, -i_Tangent, i_Normal );
	oEyeVector = mul( vCameraDirection, T );
#endif

#if DISABLE_BUMP_AND_PARALLAX
	// Set objecct-space normal vector
	oNormalOS = i_Normal;
#endif
        
#if ENVIRONMENT
	// Compute reflected vector from position to eye in tangent space
	oReflectedVectorWS = mul( i_Position.xyz - eyePosition, rotation );
#endif

#if LIGHTMAP
	oTextureCoord.zw = lighmapTexCoord.xy;
#endif 

#if FOG
	// Compute pixel depth
    oPixelDepth = oPosition.z / g_vFogParams[ FOG_CAMERA_FAR_CLIP_PLANE ];
#endif

#if OUTPUT_DEPTH
	// Compute pixel depth
	oLightDirectionAndDepth[ DEPTH_INDEX ] = oPosition.z / g_fCameraFarPlaneDist;
#endif
}

void PS_Factory(
#if !LIGHTMAP
#if !USE_MATERIAL_COLOR
					float2 TexCoord							: TEXCOORD0,
#endif
#else
					float4 TexCoord							: TEXCOORD0,
#endif
#if !OUTPUT_DEPTH
					float3 lightDirAndDepth					: TEXCOORD1, // Tangent/object space
#else
					float4 lightDirAndDepth					: TEXCOORD1, // Tangent/object space
#endif
#if !DISABLE_BUMP_AND_PARALLAX
					float3 HalfVector 						: TEXCOORD2, // Tangent space
#if !DISABLE_PARALLAX
					float3 EyeVector						: TEXCOORD3, // Tangent space (for parallax)
#endif
#else
					float3 i_Normal							: TEXCOORD2, // Object space
#endif
#if ENVIRONMENT
					float3 ReflectedVectorWS				: TEXCOORD4, // World space reflected vector
#endif
#if FOG
					float i_PixelDepth						: TEXCOORD5,
#endif
					out float4 oColor 						: COLOR0	// Output color (render target 0)
#if OUTPUT_DEPTH
					, out float4 o_Depth					: COLOR1	// Output color (render target 1)
#endif
					)
{
    
#if !LIGHTMAP_ONLY
    
	// Normalizes light vector
	float3 light = normalize( lightDirAndDepth.xyz );

#if !DISABLE_BUMP_AND_PARALLAX 
	float3  H = normalize( HalfVector.xyz );
    float4 vHeightAndEnvTexture = tex2D( HeightTextureSampler, TexCoord.xy );
	
#if !DISABLE_PARALLAX
	// Do parallax mapping
    float  fHeight = vHeightAndEnvTexture.r;
    fHeight = 1.0f - fHeight;
    fHeight = fHeight * 0.03f - 0.015f;
    EyeVector = normalize( EyeVector);
    TexCoord.xy = TexCoord.xy + float2(EyeVector.y, EyeVector.x) * fHeight;
#endif	
    
	// Sample and expand the normal map texture	
	float4 vNormalTex = tex2D( BumpTextureSampler, TexCoord );
    float3 normal = vNormalTex.rgb;
	normal = expand( normal );
#else
	float3 normal = i_Normal;
#endif

#if !USE_MATERIAL_COLOR
	// Get diffuse color from basetexture	
	float4 vDiffuseColor = tex2D( BaseTextureSampler, TexCoord );
#else
	// Get diffuse color from material
	float4 vDiffuseColor = g_MaterialColor;
#endif

#if ENVIRONMENT
	// Get environment reflected color (reflect vector from eye to position and normal in tangent space)
	float3 vEnvTexCoord = reflect( ReflectedVectorWS, normal );
	vDiffuseColor.rgb = lerp( vDiffuseColor.rgb, texCUBE( EnvironmentMapSampler, vEnvTexCoord ).rgb, vHeightAndEnvTexture.g );
#endif      

	// Add simple N.L shading
	oColor.rgb = vDiffuseColor.rgb * ( max( dot(normal,light), 0.4f) );

#if !DISABLE_BUMP_AND_PARALLAX  
	// Add specular
	oColor.rgb += ( pow( dot( normal, H ), 20.0f ) * vNormalTex.a );
#endif    
       
#if LIGHTMAP
	// Add lightmap color
	oColor.rgb *= ( tex2D( LightMapSampler, TexCoord.zw ) * g_vLightMapParam.x + g_vLightMapParam.y );
#endif

	// Set glow strength
	oColor.a = vDiffuseColor.a;

#if !USE_MATERIAL_COLOR	
	// Increase diffuse color by glow
	oColor.rgb += ( vDiffuseColor.rgb * vDiffuseColor.a );
#endif	
    
#if FOG
	// Compute distance fog intensity in the screen space
	float fFogIntensity;
	fFogIntensity = i_PixelDepth;
	fFogIntensity = ( fFogIntensity - g_vFogParams[ FOG_NEAR_PLANE ] ) / ( g_vFogParams[ FOG_FAR_PLANE ] - g_vFogParams[ FOG_NEAR_PLANE ] );
	fFogIntensity = clamp( fFogIntensity, 0.0f, 1.0f );    
	
	// Add fog
	oColor = lerp( oColor, g_vFogColor, fFogIntensity );
#endif    

#if OUTPUT_DEPTH
	o_Depth = ComputeDepth( lightDirAndDepth[ DEPTH_INDEX ] );
	//oColor = o_Depth;
	//oColor.a = 0;
#endif

#else // LIGHTMAP_ONLY

    // Add lightmap color only
    oColor.rgb = tex2D( LightMapSampler, TexCoord.zw );
    oColor.a = 0;
#endif    
}

//**********************************************************************************************
// Cyborg
//**********************************************************************************************

void VS_Cyborg(	float4 i_Position						: POSITION,
					float3 i_Normal							: NORMAL,
#if SKIN
					float4 i_BlendWeights					: BLENDWEIGHT,
					float4 i_BlendIndices					: BLENDINDICES,
#endif
					float2 i_LighmapTexCoord				: TEXCOORD0,
					out float4 o_Position					: POSITION,
					out float2 o_LightmapTexCoord			: TEXCOORD0,
					out float3 o_Normal						: TEXCOORD1,
					out float3 o_LightDirection				: TEXCOORD2,
					out float3 o_HalfVector					: TEXCOORD3,
					out float3 o_ReflectedVector			: TEXCOORD4 )
{
#if !SKIN
    
	// Transform the position from object space to homogeneous projection space
	o_Position = mul( i_Position, g_mWorldViewProjection );

#else
    
	// NOTE: This shader is invalid, because it doesn't compute proper normal vector and object space position,
	//       but in this case, it is not a problem - skinned mesh is very thin and it looks good even with this "bug"
    
	float4		vPositionWorld	= 0.0f;
	float		LastWeight		= 0.0f;
     
	// Cast the vectors to arrays for use in the for loop below
	float	BlendWeightsArray[4]	= (float[4])i_BlendWeights;  
	int		IndexArray[4]			= (int[4])i_BlendIndices;
/*
	// Calculate the pos/normal using the "normal" weights 
	// and accumulate the weights to calculate the last weight
	for (int iBone = 0; iBone < NUM_OF_BONES-1; iBone++)
	{
		LastWeight = LastWeight + BlendWeightsArray[iBone];
		vPositionWorld += mul( i_Position, g_mWorldMatrixArray[IndexArray[iBone]] ) * BlendWeightsArray[iBone];
	}
	LastWeight = 1.0f - LastWeight; 

	// Now that we have the calculated weight, add in the final influence
	vPositionWorld += ( mul(i_Position, g_mWorldMatrixArray[IndexArray[NUM_OF_BONES-1]]) * LastWeight );
    */
    vPositionWorld += ( mul(i_Position, g_mWorldMatrixArray[IndexArray[0]]) * BlendWeightsArray[0] );
    vPositionWorld += ( mul(i_Position, g_mWorldMatrixArray[IndexArray[1]]) * BlendWeightsArray[1]);
    //vPositionWorld += ( mul(i_Position, g_mWorldMatrixArray[IndexArray[2]]) * BlendWeightsArray[2]);
    //vPositionWorld += ( mul(i_Position, g_mWorldMatrixArray[IndexArray[3]]) * BlendWeightsArray[3]);
    
    LastWeight = i_BlendWeights[0] + BlendWeightsArray[1];// + BlendWeightsArray[2]+ BlendWeightsArray[3];
    

	// Transform position from world space into view and then projection space
	o_Position = mul( vPositionWorld, g_mViewProj );

    o_Position = lerp( mul( i_Position, g_mWorldViewProjection ), o_Position, LastWeight );

#endif        

	// Lightmap texture coordinates
	o_LightmapTexCoord.xy = i_LighmapTexCoord;    
	 
	// Set object space normal vector
#if COMPRESSED_VECTORS	
	o_Normal = expand( i_Normal );
#else
	o_Normal = i_Normal;
#endif
    
	// Compute the object-space light vector
	o_LightDirection = g_LightPositionOS - i_Position.xyz;
    
	// Compute half vector
	float3 vCameraDirection = g_EyePositionOS - i_Position.xyz;
	o_HalfVector.xyz = o_LightDirection + vCameraDirection;
    
	// Compute reflected vector from position of vertex to the camera in object space
	o_ReflectedVector = i_Position.xyz - g_EyePositionOS;
}

void PS_Cyborg( 	
					float2 i_LightmapTexCoord		: TEXCOORD0,
					float3 i_Normal					: TEXCOORD1,
					float3 i_LightDirection			: TEXCOORD2,
					float3 i_HalfVector				: TEXCOORD3,
					float3 i_ReflectedVector		: TEXCOORD4,
					out float4 oColor 				: COLOR,
					uniform float4 u_DiffuseColor,
					uniform float u_AmbientColor,
					uniform bool u_AddSpecular )
{
	// Set basic diffuse color
	oColor = u_DiffuseColor;
	
#if CYBORG_EYES
	oColor.rgb = tex2D( BaseTextureSampler, i_LightmapTexCoord.xy ).rgb;
#endif		
    
	// Normalize vectors
	float3 vNormal = normalize( i_Normal );
	float3 vLightDirection = normalize( i_LightDirection );
	float3 vHalf = normalize( i_HalfVector );
    
	// Diffuse lighting
	oColor.rgb *= ( dot( vNormal, vLightDirection ) + u_AmbientColor );
    
#if ENVIRONMENT
	// Sample environment reflected color (object space)
	float3 vEnvTexCoord = reflect( i_ReflectedVector, vNormal );
	float3 vEnvMapColor = texCUBE( EnvironmentMapSampler, vEnvTexCoord );
	oColor.rgb = lerp( oColor.rgb, vEnvMapColor, 0.2f );
#endif
    
#if !CYBORG_EYES
	// Add specular
	if ( u_AddSpecular == true )
	{
		oColor.rgb += pow( dot( vNormal, vHalf ), 40.0f );
	}
#endif
    
#if LIGHTMAP
    // Add color from lightmap
	oColor.rgb *= ( tex2D( LightMapSampler, i_LightmapTexCoord.xy ).rgb + 0.2f );
#endif   
}

//**********************************************************************************************
// Chrome
//**********************************************************************************************

void VS_Chrome(	float4 i_Position						: POSITION,
					float3 i_Normal							: NORMAL,
					out float4 o_Position					: POSITION,
					out float3 o_Normal						: TEXCOORD0,
					out float3 o_ReflectedVector			: TEXCOORD1 )
{
	// Transform the position from object space to homogeneous projection space
	o_Position = mul( i_Position, g_mWorldViewProjection );

#if COMPRESSED_VECTORS
	// Unpack normal vector
	i_Normal = expand( i_Normal );	
#endif	

    // Set object space normal vector
	o_Normal = i_Normal;
    
	// Compute reflected vector from position of vertex to the camera in object space
	o_ReflectedVector = -i_Position.xyz + g_EyePositionOS;
}

void PS_Chrome( 	float3 i_Normal					: TEXCOORD0,
					float3 i_ReflectedVector		: TEXCOORD1,
					out float4 oColor 				: COLOR )
{   
#if ENVIRONMENT    
	// Sample environment reflected color (object space)
	float3 vEnvTexCoord = reflect( i_ReflectedVector, i_Normal );
	oColor = float4( texCUBE( EnvironmentMapSampler, vEnvTexCoord ).rgb, 0.0f );
#else
	oColor = 0;
#endif
}

//**********************************************************************************************
// Techniques
//**********************************************************************************************
technique FactoryShader
{
	pass P0
	{
		VertexShader	= compile vs_1_1 VS_Factory( g_LightPositionOS, g_EyePositionOS, g_mWorldViewProjection );
		PixelShader		= compile ps_2_0 PS_Factory();
    }
}

technique CyborgShader
{
	pass P0
	{
		VertexShader	= compile vs_1_1 VS_Cyborg();
		PixelShader		= compile ps_2_0 PS_Cyborg( float4( 1.0f, 1.0f, 1.0f, 0.0f ), 0.2f, true );
    }
}

technique CyborgShader_Wires
{
	pass P0
	{
		VertexShader	= compile vs_1_1 VS_Cyborg();
		PixelShader		= compile ps_2_0 PS_Cyborg( float4( 1.0f, 0.0f, 0.0f, 0.0f ), 0.0f, false );
    }
}

technique CyborgShader_Chrome
{
	pass P0
	{
		VertexShader	= compile vs_1_1 VS_Chrome();
		PixelShader		= compile ps_2_0 PS_Chrome();
    }
}