//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
float4x4	g_mWorld;												// World matrix for object
float4x4	g_mWorldViewProjection;									// World * View * Projection matrix
float3		g_vCameraPosition;
float		g_fTime;
float		g_fMergeFactor			= 1.0f;

texture		g_BumpTexture;
textureCUBE	g_EnvironmentTex;

//-----------------------------------------------------------------------------
// Texture sampler
//-----------------------------------------------------------------------------
sampler g_BumpTexSampler = sampler_state
{
	Texture		= < g_BumpTexture >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
};

sampler g_EnvironmentTexSampler = sampler_state
{
	Texture		= < g_EnvironmentTex >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
	AddressW	= WRAP;
};

//-----------------------------------------------------------------------------
// Name: Fresnel
//-----------------------------------------------------------------------------
float Fresnel( float3 I, float3 Normal )
{
	float fFresnel =  0.1f + pow( 1 + dot( I, Normal ), 1.0f );
   
	// I need to bound fresnel factor to flickering from refracted image
	if ( fFresnel > 0.58f )
		fFresnel = 1;
    
	return fFresnel;
}

//-----------------------------------------------------------------------------
// Name: expand
//-----------------------------------------------------------------------------
float3 expand(float3 v)
{
	return (v-0.5)*2;
}

//-----------------------------------------------------------------------------
// Name: VS_Glass
//-----------------------------------------------------------------------------
void VS_Glass( float4		iPosition		: POSITION, 
			   float3		iNormal			: NORMAL,
			   float3 		iTangent		: TANGENT,
			   float2		iTexCoord		: TEXCOORD0,
			   out float4   oPosition		: POSITION,
			   out float3	oNormalWorld	: TEXCOORD0,
			   out float2	oBumpTexCoord1	: TEXCOORD1,
			   out float2	oBumpTexCoord2	: TEXCOORD2,
               out float3   oVertexToCamera	: TEXCOORD3,
			   out float3 	oTangentWorld	: TEXCOORD4 )
{

	// Transform the position from object space to homogeneous projection space
	oPosition = mul( iPosition, g_mWorldViewProjection);

	// Transform position normal and tangent into world-space
	float4 vPositionWorld   = mul( iPosition, g_mWorld );
	oNormalWorld	        = mul( expand( iNormal ), (float3x3)g_mWorld );
	oTangentWorld           = mul( expand( iTangent ),(float3x3)g_mWorld );
    
    // Compute vertex to camera vector
	oVertexToCamera = vPositionWorld - g_vCameraPosition;
	
	// Compute bump wave 1
	float2 vWave = ( iTexCoord.xy - float2( 0.5f, 0.5f ) ) ;
	float fSin = sin( g_fTime * 0.1f );
	float fCos = cos( g_fTime * 0.1f );
	oBumpTexCoord1.x = vWave.x * fCos - vWave.y * fSin;
	oBumpTexCoord1.y = vWave.x * fSin + vWave.y * fCos;
	
	// Compute bump wave 2
	vWave = ( iTexCoord.xy - float2( 0.5f, 0.5f ) );
	fSin = sin( -g_fTime * 0.15f );
	fCos = cos( -g_fTime * 0.15f );	
	oBumpTexCoord2.x = vWave.x * fCos - vWave.y * fSin;
	oBumpTexCoord2.y = vWave.x * fSin + vWave.y * fCos;
}

//-----------------------------------------------------------------------------
// Name: PS_Glass
//-----------------------------------------------------------------------------
void PS_Glass( float3 iNormalWorld		: TEXCOORD0,
			   float2 iBumpTexCoord1	: TEXCOORD1,
			   float2 iBumpTexCoord2	: TEXCOORD2,
			   float3 iVertexToCamera	: TEXCOORD3,
               float3 	iTangentWorld	: TEXCOORD4,
			   out float4 oColor		: COLOR,
			   uniform float etaRatio )
{
	// Sample normal map 1
	float3 vNormalBump1 = tex2D( g_BumpTexSampler, iBumpTexCoord1 );
	vNormalBump1 = expand( vNormalBump1 );
	
    // Sample normal map 2
	float3 vNormalBump2 = tex2D( g_BumpTexSampler, iBumpTexCoord2 );
	vNormalBump2 = expand( vNormalBump2 );
	
    // Compute final normal vector
	float3 vNormalBump = ( vNormalBump1 + vNormalBump2 ) * 0.5f;
    
    // Contruct matrix and transform normal vector from texture space to world space
    float3 vBinormal   = cross( iNormalWorld, iTangentWorld );
    float3x3 mRotation = transpose( float3x3( iTangentWorld, vBinormal, iNormalWorld ) );
	vNormalBump = normalize( mul( vNormalBump, mRotation ) );
	
    // Normalize vertex to camera vector
	float3 I = normalize( iVertexToCamera );

	// Bump mapping reflection/refracion
	float fFresnelFactorBump = Fresnel( I, vNormalBump );					// Fresnel term
    float3 vNormalBumpReflect = lerp( vNormalBump, iNormalWorld, 0.9f );	// Make bumped reflection softer
	float3 vReflectBump = reflect( I, vNormalBumpReflect );					// Reflection
	float3 vRefractBump = refract( I, vNormalBump, etaRatio );				// Refraction
	float4 reflectedColorBump = texCUBE( g_EnvironmentTexSampler, vReflectBump );
	float4 refractedColorBump = texCUBE( g_EnvironmentTexSampler, vRefractBump );
	float4 vColorBump = lerp( refractedColorBump, reflectedColorBump, fFresnelFactorBump );

    // Bump mapping reflection/refracion
    float fFresnelFactorNormal = Fresnel( I, iNormalWorld );
	float3 vReflectNormal = reflect( I, iNormalWorld );
	float3 vRefractNormal = refract( I, iNormalWorld, etaRatio );
	float4 reflectedColorNormal = texCUBE( g_EnvironmentTexSampler, vReflectNormal );
	float4 refractedColorNormal = texCUBE( g_EnvironmentTexSampler, vRefractNormal );
	float4 ColorNormal = lerp( refractedColorNormal, reflectedColorNormal, fFresnelFactorNormal );
    
	// Compute final color
	oColor = lerp( vColorBump, ColorNormal, g_fMergeFactor );
}

//-----------------------------------------------------------------------------
// Name: Glass
//-----------------------------------------------------------------------------
technique Glass
{
	pass P0
	{
		VertexShader	= compile vs_1_1 VS_Glass();
		PixelShader		= compile ps_2_0 PS_Glass( 1.19f ); 	// Refraction ratio
	}
}
