//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
float4x4	g_mWorldViewProjection;									// World * View * Projection matrix
float4x4	g_mWorld;
float3		g_fCameraDirection	= float3( 1.0f, 0.0f, 0.0f );
float		g_fTime		        = 0.0f;								// The time
float4		g_vWaterPlane		= float4( 0.0f, 1.0f, 0.0f, 0.0f );	// Water plane equation

texture		g_ReflectTextrure;
texture     g_RefractionTextrure;
texture		g_WaveNormalTexture;
texture		g_UnderWaterTex;

sampler WaveNormalTextureSampler = sampler_state
{
	Texture		= < g_WaveNormalTexture >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
};

sampler ReflectTextureSampler = sampler_state
{
	Texture		= < g_ReflectTextrure >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= CLAMP;
	AddressV	= CLAMP;
};

sampler RefractTextrureSampler = sampler_state
{
	Texture		= < g_RefractionTextrure >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= CLAMP;
	AddressV	= CLAMP;
};

sampler UnderWaterTexSampler = sampler_state
{
	Texture		= < g_UnderWaterTex >;
	MinFilter	= LINEAR;  
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;
	AddressV	= WRAP;
};

//-----------------------------------------------------------------------------
// Convert pixel from normal texture to normal vector
//-----------------------------------------------------------------------------
float3 Expand( float3 v )
{
	return ( v - 0.5f ) * 2;
}

//-----------------------------------------------------------------------------
// Compute Fresnel term
//-----------------------------------------------------------------------------
half Fresnel( half NdotL, half fresnelBias, half fresnelScale, half fresnelPower )
{
	return min( fresnelBias + fresnelScale * pow( 1 - NdotL, fresnelPower ), 1.0f );
}

#define NUM_OF_WAVES 2
#define BUMP_WAVE_INTENSITY 7

float g_fAmplitude[NUM_OF_WAVES] = { 1.8f, 1.3f };
float2 g_vDirection[NUM_OF_WAVES] = { float2( -0.2f, -0.8f ), float2( 0.5f, -0.7f ) };
float g_fWaceLength[NUM_OF_WAVES] = { 50.0f, 30.0f };
float g_fSpeed[NUM_OF_WAVES] = { 50.0f, 30.0f };

float GetHeight( float4 a_vPos )
{
    float fSum = 0;
    for ( int i = 0; i < NUM_OF_WAVES; i++ )
    {
        float fOmega = 6.283185f/g_fWaceLength[i];
        float fPhi = g_fSpeed[i] * fOmega;
        fSum += g_fAmplitude[i] * sin( dot( g_vDirection[i],  a_vPos.xz ) * fOmega + g_fTime * fPhi );
    }
    return fSum;
}

float GetHeightDer( float4 a_vPos )
{
    float fSum = 0;
    for ( int i = 0; i < NUM_OF_WAVES; i++ )
    {
        float fOmega = 6.283185f/g_fWaceLength[i];
        float fPhi = g_fSpeed[i] * fOmega;        
        fSum += ( g_fAmplitude[i] * BUMP_WAVE_INTENSITY * cos( dot( g_vDirection[i],  a_vPos.xz ) * fOmega + g_fTime * fPhi ) * fOmega * g_vDirection[i].x );
    }
    return fSum;
}

//-----------------------------------------------------------------------------
// Water
//-----------------------------------------------------------------------------
void VS_Water( float4	  i_Position		: POSITION, 
			   float4	  i_WaveTexCoord	: TEXCOORD0, 
			   out float2 o_Wave0	        : TEXCOORD0,
               out float2 o_Wave1	        : TEXCOORD1,
               out float4 o_PositionLocal	: TEXCOORD2,
               out float3 o_CameraDir	    : TEXCOORD3,
			   out float4 o_Position		: POSITION )
{
    // Set Y position - wave animation
    i_Position.y = GetHeight( i_Position );
    
    // Computer binormal, tangent and normal vectors for the waves
    float fFactor = GetHeightDer( i_Position );       
    float3 vBinormal = normalize( float3(1, 0, fFactor ) );
    float3 vTangent = normalize( float3( 0, 1, fFactor ) );
    float3 vNormal = cross( vBinormal, vTangent );
    
	// Construct object space to texture space 3x3 matrix
	float3x3 mRotation = transpose( float3x3(vTangent, vBinormal, vNormal) ); 
    
    // Conver camera direction vector
    o_CameraDir = mul( g_fCameraDirection, mRotation );
        
    // Transform the position from object space to homogeneous projection space
	o_Position = mul( i_Position, g_mWorldViewProjection);
    o_PositionLocal = i_Position;
        
#if 0    
    // Compute wave 0
	/*float2 vWave = i_WaveTexCoord.xy * 0.85f;
	float fSin = sin( g_fTime * 0.1f );
	float fCos = cos( g_fTime * 0.1f );
	o_Wave0.x = vWave.x * fCos - vWave.y * fSin;
	o_Wave0.y = vWave.x * fSin + vWave.y * fCos;
    
	// Compute wave 1	
	vWave = i_WaveTexCoord.xy * 1.15f;
	fSin = sin( -g_fTime * 0.2f );
	fCos = cos( g_fTime * 0.3f );	
	o_Wave1.x = vWave.x * fCos - vWave.y * fSin;
	o_Wave1.y = vWave.x * fSin + vWave.y * fCos;*/
#endif
    
    // Compute wave 0
    float2 vWave = i_WaveTexCoord.xy * 0.85f;
	o_Wave0.x = vWave.x * 2  + sin( g_fTime * 1.5f ) * 0.1f;
	o_Wave0.y = vWave.y * 2 + g_fTime*0.5f;
    
	// Compute wave 1	
	vWave = i_WaveTexCoord.xy * 1.15f;
	o_Wave1.x = vWave.x * 3 + sin( g_fTime ) * 0.5f;
	o_Wave1.y = vWave.y + g_fTime*0.1f;
}

void PS_Water( float2 i_Wave0	        : TEXCOORD0,
               float2 i_Wave1			: TEXCOORD1,
			   float4 i_PositionLocal	: TEXCOORD2,
               float3 i_CameraDir	    : TEXCOORD3,
               out float4 o_Color		: COLOR )
{
    float2 vTexCoord;
    float4 vPositionScreen = mul( i_PositionLocal, g_mWorldViewProjection);
    vTexCoord = vPositionScreen.xy / vPositionScreen.w;
    vTexCoord.y *= -1.0f;
    vTexCoord = ( vTexCoord + 1 ) / 2;
    
	// Get bump layer
	float3 vBumpTex0 = tex2D( WaveNormalTextureSampler, i_Wave0 ).xyz;
	float3 vBumpTex1 = tex2D( WaveNormalTextureSampler, i_Wave1 ).xyz;
	
    // Compute average bump vector
	float3 vBumpTex = ( vBumpTex0 + vBumpTex1 ) / 2.0f;
    
    // Convert bump to normal vector
	half3 vBumpNormal = Expand( vBumpTex.xyz );
    
    // Apply individual bump scale for refraction and reflection
    half3 vReflBump = vBumpNormal.xyz * half3( 0.1, 0.1, 1.0 );
    half3 vRefrBump = vBumpNormal.xyz * half3( 0.05, 0.05, 1.0 );
    
	// Sample refrlected and refracted textures
	float4 vReflColor = tex2D( ReflectTextureSampler, vTexCoord.xy + vReflBump.xy );
	float4 vRefrColor = tex2D( RefractTextrureSampler, vTexCoord.xy + vRefrBump.xy );
    
    half4 vUnderWater = tex2D( UnderWaterTexSampler, vTexCoord.xy );
	vRefrColor.rgb += vUnderWater.rgb * 0.5f;	// Add caustics (underwater reflections)
	
   	// Compute Fresnel term
    float NdotL = dot( normalize( i_CameraDir ), vBumpNormal );
	float fFresnel = Fresnel( NdotL, 0.1f, 0.55f, 1.0f );    
    
	// Set final color
	o_Color = lerp( vRefrColor, vReflColor, fFresnel );

    // Modulate output color
    o_Color.rgba *= float4( 0.74f, 0.91f, 1.0f, 10.0f );
}

//-----------------------------------------------------------------------------
// DeepWater
//-----------------------------------------------------------------------------
void VS_DeepWater( float4 i_Position			: POSITION,
				   float3 i_Normal				: NORMAL,
				   out float4 o_Position		: POSITION,
				   out float3 o_Distance		: TEXCOORD0 )
{
	// Set position
	o_Position = mul( i_Position, g_mWorldViewProjection );	
    
    float3 vNormalWorld = mul( i_Normal, (float3x3)g_mWorld );
	
	// Compute world position and distance from plane
	float3 vPositionWorld = mul( i_Position, g_mWorld );
	o_Distance.xy = vPositionWorld.xz / 50.0f;
	o_Distance.z = max( 0.0f, dot( -g_vWaterPlane.xyz, vNormalWorld.xyz ) );
}

void PS_DeepWater( float3 i_Distance			: TEXCOORD0,
				   out float4 o_Color			: COLOR )
{   
	half3 vUnderWaterTex = tex2D( UnderWaterTexSampler, i_Distance.xy );
	vUnderWaterTex *= i_Distance.z;
	o_Color = float4( vUnderWaterTex.xyz, 1.0f );
}

//-----------------------------------------------------------------------------
// Techniques
//-----------------------------------------------------------------------------

technique Water
{
	pass P0
	{
       //FillMode = WIREFRAME;
		VertexShader	= compile vs_1_1 VS_Water();
		PixelShader		= compile ps_2_0 PS_Water();
	}
}

// Deep water
technique DeepWater
{
	pass P0
	{
		VertexShader	= compile vs_1_1 VS_DeepWater();
		PixelShader		= compile ps_2_0 PS_DeepWater();
	}
}
